"""Module to contain models for hpc rest api responses."""

import json
from pathlib import PurePosixPath
from typing import Any, List, Literal, Optional

from pydantic import Extra, Field, SecretStr
from pydantic_settings import BaseSettings


class HpcJobSubmitSettings(BaseSettings):
    """Configs to send in a post request. v0.0.36 of slurm rest api."""

    account: Optional[str] = Field(
        None,
        description="Charge resources used by this job to specified account.",
    )
    # This is the way it is written in the slurm documentation
    account_gather_freqency: Optional[str] = Field(
        None,
        description=(
            "Define the job accounting and profiling sampling intervals."
        ),
    )
    argv: Optional[List[str]] = Field(
        None, description="Arguments to the script."
    )
    array: Optional[str] = Field(
        None,
        description=(
            "Submit a job array, multiple jobs to be executed with identical "
            "parameters. The indexes specification identifies what array "
            "index values should be used."
        ),
    )
    batch_features: Optional[str] = Field(
        None, description="features required for batch script's node"
    )
    begin_time: Optional[str] = Field(
        None,
        description=(
            "Submit the batch script to the Slurm controller immediately, "
            "like normal, but tell the controller to defer the allocation of "
            "the job until the specified time."
        ),
    )
    burst_buffer: Optional[str] = Field(
        None, description="Burst buffer specification."
    )
    cluster_constraints: Optional[str] = Field(
        None,
        description=(
            "Specifies features that a federated cluster must have to have a "
            "sibling job submitted to it."
        ),
    )
    comment: Optional[str] = Field(None, description="An arbitrary comment.")
    constraints: Optional[str] = Field(
        None, description="node features required by job."
    )
    core_specification: Optional[int] = Field(
        None,
        description=(
            "Count of specialized threads per node reserved by the job for "
            "system operations and not used by the application."
        ),
    )
    cores_per_socket: Optional[int] = Field(
        None,
        description=(
            "Restrict node selection to nodes with at least the specified "
            "number of cores per socket."
        ),
    )
    cpu_binding: Optional[str] = Field(None, description="Cpu binding")
    cpu_binding_hint: Optional[str] = Field(
        None, description="Cpu binding hint"
    )
    cpu_frequency: Optional[str] = Field(
        None,
        description=(
            "Request that job steps initiated by srun commands inside this "
            "sbatch script be run at some requested frequency if possible, on "
            "the CPUs selected for the step on the compute node(s)."
        ),
    )
    cpus_per_gpu: Optional[str] = Field(
        None, description="Number of CPUs requested per allocated GPU."
    )
    cpus_per_task: Optional[int] = Field(
        None,
        description=(
            "Advise the Slurm controller that ensuing job steps will require "
            "ncpus number of processors per task."
        ),
    )
    current_working_directory: Optional[str] = Field(
        None,
        description=(
            "Instruct Slurm to connect the batch script's standard output "
            "directly to the file name."
        ),
    )
    deadline: Optional[str] = Field(
        None,
        description=(
            "Remove the job if no ending is possible before this deadline "
            "(start > (deadline - time[-min]))."
        ),
    )
    delay_boot: Optional[int] = Field(
        None,
        description=(
            "Do not reboot nodes in order to satisfied this job's feature "
            "specification if the job has been eligible to run for less than "
            "this time period."
        ),
    )
    dependency: Optional[str] = Field(
        None,
        description=(
            "Defer the start of this job until the specified dependencies "
            "have been satisfied completed."
        ),
    )
    distribution: Optional[str] = Field(
        None,
        description=(
            "Specify alternate distribution methods for remote processes."
        ),
    )
    environment: Optional[dict] = Field(
        None, description="Dictionary of environment entries."
    )
    exclusive: Optional[Literal["user", "mcs", "true", "false"]] = Field(
        None,
        description=(
            "The job allocation can share nodes just other users with the "
            "'user' option or with the 'mcs' option)."
        ),
    )
    get_user_environment: Optional[bool] = Field(
        None, description="Load new login environment for user on job node."
    )
    gres: Optional[str] = Field(
        None,
        description=(
            "Specifies a comma delimited list of generic consumable resources."
        ),
    )
    gres_flags: Optional[
        Literal["disable-binding", "enforce-binding"]
    ] = Field(
        None, description="Specify generic resource task binding options."
    )
    gpu_binding: Optional[str] = Field(
        None, description="Requested binding of tasks to GPU."
    )
    gpu_frequency: Optional[str] = Field(
        None, description="Requested GPU frequency."
    )
    gpus: Optional[str] = Field(None, description="GPUs per job.")
    gpus_per_node: Optional[str] = Field(None, description="GPUs per node.")
    gpus_per_socket: Optional[str] = Field(
        None, description="GPUs per socket."
    )
    gpus_per_task: Optional[str] = Field(None, description="GPUs per task.")
    hold: Optional[bool] = Field(
        None,
        description=(
            "Specify the job is to be submitted in a held state "
            "(priority of zero)."
        ),
    )
    kill_on_invalid_dependency: Optional[bool] = Field(
        None,
        description=(
            "If a job has an invalid dependency, then Slurm is to "
            "terminate it."
        ),
    )
    licenses: Optional[str] = Field(
        None,
        description=(
            "Specification of licenses (or other resources available on all "
            "nodes of the cluster) which must be allocated to this job."
        ),
    )
    mail_type: Optional[str] = Field(
        None,
        description="Notify user by email when certain event types occur.",
    )
    mail_user: Optional[str] = Field(
        None,
        description=(
            "User to receive email notification of state changes as defined "
            "by mail_type."
        ),
    )
    mcs_label: Optional[str] = Field(
        None,
        description="This parameter is a group among the groups of the user.",
    )
    memory_binding: Optional[str] = Field(
        None, description="Bind tasks to memory."
    )
    memory_per_cpu: Optional[int] = Field(
        None, description="Minimum real memory per cpu (MB)."
    )
    memory_per_gpu: Optional[int] = Field(
        None, description="Minimum memory required per allocated GPU."
    )
    memory_per_node: Optional[int] = Field(
        None, description="Minimum real memory per node (MB)."
    )
    minimum_cpus_per_node: Optional[int] = Field(
        None, description="Minimum number of CPUs per node."
    )
    minimum_nodes: Optional[bool] = Field(
        None,
        description=(
            "If a range of node counts is given, prefer the smaller count."
        ),
    )
    name: Optional[str] = Field(
        None, description="Specify a name for the job allocation."
    )
    nice: Optional[str] = Field(
        None,
        description=(
            "Run the job with an adjusted scheduling priority within Slurm."
        ),
    )
    no_kill: Optional[bool] = Field(
        None,
        description=(
            "Do not automatically terminate a job if one of the nodes it has "
            "been allocated fails."
        ),
    )
    nodes: Optional[List[int]] = Field(
        None,
        description=(
            "Request that a minimum of minnodes nodes and a maximum node "
            "count."
        ),
    )
    open_mode: Optional[Literal["append", "truncate"]] = Field(
        None,
        description=(
            "Open the output and error files using append or truncate mode "
            "as specified."
        ),
    )
    partition: Optional[str] = Field(
        None,
        description=(
            "Request a specific partition for the resource allocation."
        ),
    )
    priority: Optional[str] = Field(
        None, description="Request a specific job priority."
    )
    # Set this to "production" for production environment
    qos: Optional[str] = Field(
        None, description="Request a quality of service for the job."
    )
    requeue: Optional[bool] = Field(
        None,
        description=(
            "Specifies that the batch job should eligible to being requeue."
        ),
    )
    reservation: Optional[str] = Field(
        None,
        description=(
            "Allocate resources for the job from the named reservation."
        ),
    )
    signal: Optional[str] = Field(
        None,
        description=(
            "When a job is within sig_time seconds of its end time, send it "
            "the signal sig_num."
        ),
    )
    sockets_per_node: Optional[int] = Field(
        None,
        description=(
            "Restrict node selection to nodes with at least the specified "
            "number of sockets."
        ),
    )
    spread_job: Optional[bool] = Field(
        None,
        description=(
            "Spread the job allocation over as many nodes as possible and "
            "attempt to evenly distribute tasks across the allocated nodes."
        ),
    )
    standard_error: Optional[str] = Field(
        None,
        description=(
            "Instruct Slurm to connect the batch script's standard error "
            "directly to the file name."
        ),
    )
    standard_in: Optional[str] = Field(
        None,
        description=(
            "Instruct Slurm to connect the batch script's standard input "
            "directly to the file name specified."
        ),
    )
    standard_out: Optional[str] = Field(
        None,
        description=(
            "Instruct Slurm to connect the batch script's standard output "
            "directly to the file name."
        ),
    )
    tasks: Optional[int] = Field(
        None,
        description=(
            "Advises the Slurm controller that job steps run within the "
            "allocation will launch a maximum of number tasks and to provide "
            "for sufficient resources."
        ),
    )
    tasks_per_core: Optional[int] = Field(
        None, description="Request the maximum ntasks be invoked on each core."
    )
    tasks_per_node: Optional[int] = Field(
        None, description="Request the maximum ntasks be invoked on each node."
    )
    tasks_per_socket: Optional[int] = Field(
        None,
        description="Request the maximum ntasks be invoked on each socket.",
    )
    thread_specification: Optional[int] = Field(
        None,
        description=(
            "Count of specialized threads per node reserved by the job for "
            "system operations and not used by the application."
        ),
    )
    threads_per_core: Optional[int] = Field(
        None,
        description=(
            "Restrict node selection to nodes with at least the specified "
            "number of threads per core."
        ),
    )
    time_limit: Optional[int] = Field(None, description="Step time limit.")
    time_minimum: Optional[int] = Field(
        None, description="Minimum run time in minutes."
    )
    wait_all_nodes: Optional[bool] = Field(
        None,
        description=(
            "Do not begin execution until all nodes are ready for " "use."
        ),
    )
    wckey: Optional[str] = Field(
        None, description="Specify wckey to be used with job."
    )

    class Config:
        """Config to set env var prefix to HPC"""

        extra = Extra.forbid
        env_prefix = "HPC_"

    @staticmethod
    def script_command_str(sif_loc_str) -> str:
        """This is the command that will be sent to the hpc"""
        command_str = [
            "#!/bin/bash",
            "\nsingularity",
            "exec",
            "--cleanenv",
            sif_loc_str,
            "python",
            "-m",
            "aind_data_transfer.jobs.basic_job",
        ]
        return " ".join(command_str)

    @staticmethod
    def _set_default_val(values: dict, key: str, default_value: Any) -> None:
        """Util method to set a default if value not in dict[key]"""
        if values.get(key) is None:
            values[key] = default_value
        return None

    @classmethod
    def from_upload_job_configs(
        cls,
        logging_directory: PurePosixPath,
        aws_secret_access_key: SecretStr,
        aws_access_key_id: str,
        aws_default_region: str,
        aws_session_token: Optional[SecretStr] = None,
        **kwargs,
    ):
        """
        Class constructor to use when submitting a basic upload job request
        Parameters
        ----------
        logging_directory : PurePosixPath
        aws_secret_access_key : SecretStr
        aws_access_key_id : str
        aws_default_region : str
        aws_session_token : Optional[SecretStr]
        kwargs : dict
          Hpc settings
        """
        hpc_env = {
            "PATH": "/bin:/usr/bin/:/usr/local/bin/",
            "LD_LIBRARY_PATH": "/lib/:/lib64/:/usr/local/lib",
            "SINGULARITYENV_AWS_SECRET_ACCESS_KEY": (
                aws_secret_access_key.get_secret_value()
            ),
            "SINGULARITYENV_AWS_ACCESS_KEY_ID": aws_access_key_id,
            "SINGULARITYENV_AWS_DEFAULT_REGION": aws_default_region,
        }
        if aws_session_token is not None:
            hpc_env[
                "SINGULARITYENV_AWS_SESSION_TOKEN"
            ] = aws_session_token.get_secret_value()
        cls._set_default_val(kwargs, "environment", hpc_env)
        cls._set_default_val(
            kwargs,
            "standard_out",
            str(logging_directory / (kwargs["name"] + ".out")),
        )
        cls._set_default_val(
            kwargs,
            "standard_error",
            str(logging_directory / (kwargs["name"] + "_error.out")),
        )
        return cls(**kwargs)

    @classmethod
    def attach_configs_to_script(
        cls,
        script: str,
        base_configs: dict,
        upload_configs_aws_param_store_name: Optional[str],
        staging_directory: Optional[str],
    ) -> str:
        """
        Helper method to attach configs to a base run command string.
        Parameters
        ----------
        script : str
          Can be like
          '#!/bin/bash \nsingularity exec --cleanenv
          feat_289.sif python -m aind_data_transfer.jobs.basic_job'
        base_configs : dict
          job_configs to attach as --json-args
        upload_configs_aws_param_store_name : Optional[str]
          Will supply this config if not in base_configs and not None
        staging_directory : Optional[str]
          Will supply this config if not in base_configs and not None

        Returns
        -------
        str
          The run command script to send to submit to the slurm cluster

        """
        if staging_directory is not None:
            cls._set_default_val(
                base_configs, "temp_directory", staging_directory
            )
        if upload_configs_aws_param_store_name is not None:
            cls._set_default_val(
                base_configs,
                "aws_param_store_name",
                upload_configs_aws_param_store_name,
            )

        return " ".join(
            [
                script,
                "--json-args",
                "'",
                json.dumps(base_configs),
                "'",
            ]
        )



"""Module for internal data models used in application"""

import ast
import os
from datetime import datetime, timedelta, timezone
from typing import ClassVar, List, Optional, Union

from aind_data_schema_models.modalities import Modality
from mypy_boto3_ssm.type_defs import ParameterMetadataTypeDef
from pydantic import AwareDatetime, BaseModel, Field, field_validator
from starlette.datastructures import QueryParams


class AirflowDagRun(BaseModel):
    """Data model for dag_run entry when requesting info from airflow"""

    conf: Optional[dict]
    dag_id: Optional[str]
    dag_run_id: Optional[str]
    data_interval_end: Optional[AwareDatetime]
    data_interval_start: Optional[AwareDatetime]
    end_date: Optional[AwareDatetime]
    execution_date: Optional[AwareDatetime]
    external_trigger: Optional[bool]
    last_scheduling_decision: Optional[AwareDatetime]
    logical_date: Optional[AwareDatetime]
    note: Optional[str]
    run_type: Optional[str]
    start_date: Optional[AwareDatetime]
    state: Optional[str]


class AirflowDagRunsResponse(BaseModel):
    """Data model for response when requesting info from dag_runs endpoint"""

    dag_runs: List[AirflowDagRun]
    total_entries: int


class AirflowDagRunsRequestParameters(BaseModel):
    """Model for parameters when requesting info from dag_runs endpoint"""

    dag_ids: list[str] = ["transform_and_upload", "transform_and_upload_v2"]
    page_limit: int = 100
    page_offset: int = 0
    states: Optional[list[str]] = []
    execution_date_gte: Optional[str] = (
        datetime.now(timezone.utc) - timedelta(weeks=2)
    ).isoformat()
    execution_date_lte: Optional[str] = None
    order_by: str = "-execution_date"

    @field_validator("execution_date_gte", mode="after")
    def validate_min_execution_date(cls, execution_date_gte: str):
        """Validate the earliest submit date filter is within 2 weeks"""
        min_execution_date = datetime.now(timezone.utc) - timedelta(weeks=2)
        # datetime.fromisoformat does not support Z in python < 3.11
        date_to_check = execution_date_gte.replace("Z", "+00:00")
        if datetime.fromisoformat(date_to_check) < min_execution_date:
            raise ValueError(
                "execution_date_gte must be within the last 2 weeks"
            )
        return execution_date_gte

    @classmethod
    def from_query_params(cls, query_params: QueryParams):
        """Maps the query parameters to the model"""
        params = dict(query_params)
        if "states" in params:
            params["states"] = ast.literal_eval(params["states"])
        return cls.model_validate(params)


class AirflowTaskInstancesRequestParameters(BaseModel):
    """Model for parameters when requesting info from task_instances
    endpoint"""

    dag_id: str = Field(..., min_length=1)
    dag_run_id: str = Field(..., min_length=1)

    @classmethod
    def from_query_params(cls, query_params: QueryParams):
        """Maps the query parameters to the model"""
        params = dict(query_params)
        return cls.model_validate(params)


class AirflowTaskInstance(BaseModel):
    """Data model for task_instance entry when requesting info from airflow"""

    dag_id: Optional[str]
    dag_run_id: Optional[str]
    duration: Optional[Union[int, float]]
    end_date: Optional[AwareDatetime]
    execution_date: Optional[AwareDatetime]
    executor_config: Optional[str]
    hostname: Optional[str]
    map_index: Optional[int]
    max_tries: Optional[int]
    note: Optional[str]
    operator: Optional[str]
    pid: Optional[int]
    pool: Optional[str]
    pool_slots: Optional[int]
    priority_weight: Optional[int]
    queue: Optional[str]
    queued_when: Optional[AwareDatetime]
    rendered_fields: Optional[dict]
    sla_miss: Optional[dict]
    start_date: Optional[AwareDatetime]
    state: Optional[str]
    task_id: Optional[str]
    trigger: Optional[dict]
    triggerer_job: Optional[dict]
    try_number: Optional[int]
    unixname: Optional[str]


class AirflowTaskInstancesResponse(BaseModel):
    """Data model for response when requesting info from task_instances
    endpoint"""

    task_instances: List[AirflowTaskInstance]
    total_entries: int


class AirflowTaskInstanceLogsRequestParameters(BaseModel):
    """Model for parameters when requesting info from task_instance_logs
    endpoint"""

    # excluded fields are used to build the url
    dag_id: str = Field(..., min_length=1, exclude=True)
    dag_run_id: str = Field(..., min_length=1, exclude=True)
    task_id: str = Field(..., min_length=1, exclude=True)
    try_number: int = Field(..., ge=0, exclude=True)
    map_index: int = Field(..., ge=-1)
    full_content: bool = True

    @classmethod
    def from_query_params(cls, query_params: QueryParams):
        """Maps the query parameters to the model"""
        params = dict(query_params)
        return cls.model_validate(params)


class JobStatus(BaseModel):
    """Model for what we want to render to the user."""

    dag_id: Optional[str] = Field(None)
    end_time: Optional[datetime] = Field(None)
    job_id: Optional[str] = Field(None)
    job_state: Optional[str] = Field(None)
    name: Optional[str] = Field(None)
    job_type: Optional[str] = Field(None)
    comment: Optional[str] = Field(None)
    start_time: Optional[datetime] = Field(None)
    submit_time: Optional[datetime] = Field(None)

    @classmethod
    def from_airflow_dag_run(cls, airflow_dag_run: AirflowDagRun):
        """Maps the fields from the HpcJobStatusResponse to this model"""
        name = airflow_dag_run.conf.get("s3_prefix", "")
        job_type = airflow_dag_run.conf.get("job_type", "")
        # v1 job_type is in CO configs
        if job_type == "":
            job_type = airflow_dag_run.conf.get("codeocean_configs", {}).get(
                "job_type", ""
            )
        return cls(
            dag_id=airflow_dag_run.dag_id,
            end_time=airflow_dag_run.end_date,
            job_id=airflow_dag_run.dag_run_id,
            job_state=airflow_dag_run.state,
            name=name,
            job_type=job_type,
            comment=airflow_dag_run.note,
            start_time=airflow_dag_run.start_date,
            submit_time=airflow_dag_run.execution_date,
        )

    @property
    def jinja_dict(self):
        """Map model to a dictionary that jinja can render"""
        return self.model_dump(exclude_none=True)


class JobTasks(BaseModel):
    """Model for what is rendered to the user for each task."""

    dag_id: Optional[str] = Field(None)
    job_id: Optional[str] = Field(None)
    task_id: Optional[str] = Field(None)
    try_number: Optional[int] = Field(None)
    task_state: Optional[str] = Field(None)
    priority_weight: Optional[int] = Field(None)
    map_index: Optional[int] = Field(None)
    submit_time: Optional[datetime] = Field(None)
    start_time: Optional[datetime] = Field(None)
    end_time: Optional[datetime] = Field(None)
    duration: Optional[Union[int, float]] = Field(None)
    comment: Optional[str] = Field(None)

    @classmethod
    def from_airflow_task_instance(
        cls, airflow_task_instance: AirflowTaskInstance
    ):
        """Maps the fields from the HpcJobStatusResponse to this model"""
        return cls(
            dag_id=airflow_task_instance.dag_id,
            job_id=airflow_task_instance.dag_run_id,
            task_id=airflow_task_instance.task_id,
            try_number=airflow_task_instance.try_number,
            task_state=airflow_task_instance.state,
            priority_weight=airflow_task_instance.priority_weight,
            map_index=airflow_task_instance.map_index,
            submit_time=airflow_task_instance.execution_date,
            start_time=airflow_task_instance.start_date,
            end_time=airflow_task_instance.end_date,
            duration=airflow_task_instance.duration,
            comment=airflow_task_instance.note,
        )


class JobParamInfo(BaseModel):
    """Model for job parameter info from AWS Parameter Store"""

    _MODALITIES_LIST: ClassVar[list[str]] = list(
        Modality.abbreviation_map.keys()
    )
    _MODALITY_TASKS: ClassVar[list[str]] = [
        "modality_transformation_settings",
        "codeocean_pipeline_settings",
    ]

    name: Optional[str]
    last_modified: Optional[datetime]
    job_type: str = Field(..., pattern=r"^[^\s/]+$")
    task_id: str = Field(..., pattern=r"^[^\s/]+$")
    modality: Optional[str]
    version: Optional[str] = Field(..., pattern=r"^(v1|v2)?$")

    @field_validator("modality", mode="after")
    def validate_modality(cls, v):
        """Check that modality is one of aind-data-schema modalities"""
        if v is not None and v not in JobParamInfo._MODALITIES_LIST:
            raise ValueError(
                "Invalid modality: modality must be one of "
                f"{JobParamInfo._MODALITIES_LIST}"
            )
        return v

    @classmethod
    def from_aws_describe_parameter(
        cls,
        parameter: ParameterMetadataTypeDef,
        job_type: str,
        task_id: str,
        modality: Optional[str],
        version: Optional[str],
    ):
        """Map the parameter to the model"""
        return cls(
            name=parameter.get("Name"),
            last_modified=parameter.get("LastModifiedDate"),
            job_type=job_type,
            task_id=task_id,
            modality=modality,
            version=version,
        )

    @staticmethod
    def get_parameter_prefix(version: Optional[str] = None) -> str:
        """Get the prefix for job_type parameters"""
        prefix = os.getenv("AIND_AIRFLOW_PARAM_PREFIX")
        if version is None or version == "v1":
            return prefix
        return f"{prefix}/{version}"

    @staticmethod
    def get_parameter_regex(version: Optional[str] = None) -> str:
        """Create the regex pattern to match the parameter name"""
        prefix = os.getenv("AIND_AIRFLOW_PARAM_PREFIX")
        regex = (
            "(?P<job_type>[^/]+)/tasks/(?P<task_id>[^/]+)"
            "(?:/(?P<modality>[^/]+))?"
        )
        if version is None or version == "v1":
            return f"{prefix}/{regex}"
        return f"{prefix}/{version}/{regex}"

    @staticmethod
    def get_parameter_name(
        job_type: str,
        task_id: str,
        modality: Optional[str],
        version: Optional[str] = None,
    ) -> str:
        """Create the parameter name from job_type and task_id"""
        prefix = JobParamInfo.get_parameter_prefix(version)
        if modality:
            return f"{prefix}/{job_type}/tasks/{task_id}/{modality}"
        return f"{prefix}/{job_type}/tasks/{task_id}"



"""Module to test configs"""

import json
import unittest
from datetime import datetime

from aind_data_schema_models.modalities import Modality
from aind_data_schema_models.platforms import Platform
from pydantic import ValidationError

from aind_data_transfer_service.models.core import (
    SubmitJobRequestV2,
    Task,
    UploadJobConfigsV2,
    validation_context,
)


class TestTask(unittest.TestCase):
    """Tests Task class"""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up test class"""
        custom_task = Task(
            image="some_image",
            image_version="1.0.0",
            image_resources={"key": "value"},
            job_settings={"param1": "value1", "param2": "value2"},
            command_script="#!/bin/bash \necho 'hello world'",
        )
        cls.custom_task = custom_task
        cls.custom_task_configs = custom_task.model_dump()

    def test_constructor(self):
        """Tests that Tasks can be created correctly."""
        # custom task
        self.assertDictEqual(
            {
                "skip_task": False,
                "image": "some_image",
                "image_version": "1.0.0",
                "image_resources": {"key": "value"},
                "job_settings": {
                    "param1": "value1",
                    "param2": "value2",
                },
                "command_script": "#!/bin/bash \necho 'hello world'",
            },
            json.loads(self.custom_task.model_dump_json()),
        )
        # skippable task
        expected_configs = {
            "skip_task": True,
            "image": None,
            "image_version": None,
            "image_resources": None,
            "job_settings": None,
            "command_script": None,
        }
        task = Task(skip_task=True)
        self.assertDictEqual(
            expected_configs, json.loads(task.model_dump_json())
        )

    def test_json_serializable(self):
        """Tests that an error is raised if dicts are not json serializable"""
        for field in [
            "image_resources",
            "job_settings",
        ]:
            expected_error = f"Value error, {field} must be json serializable!"
            invalid_configs = self.custom_task_configs.copy()
            invalid_configs[field] = {"param1": list}
            with self.assertRaises(ValidationError) as e:
                Task(**invalid_configs)
            errors = e.exception.errors()
            self.assertEqual(1, len(errors))
            self.assertIn(expected_error, errors[0]["msg"])


class TestUploadJobConfigsV2(unittest.TestCase):
    """Tests UploadJobConfigsV2 class"""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up test class"""
        example_configs = UploadJobConfigsV2(
            job_type="default",
            project_name="Behavior Platform",
            platform=Platform.BEHAVIOR,
            modalities=[Modality.BEHAVIOR_VIDEOS],
            subject_id="123456",
            acq_datetime=datetime(2020, 10, 13, 13, 10, 10),
            tasks={
                "modality_transformation_settings": {
                    "behavior-videos": Task(
                        job_settings={"input_source": "dir/data_set_1"},
                    ),
                }
            },
        )

        cls.example_configs = example_configs
        cls.base_configs = example_configs.model_dump(
            exclude={"s3_bucket": True, "s3_prefix": True}
        )

    def test_s3_prefix(self):
        """Test s3_prefix computed property"""
        self.assertEqual(
            "behavior_123456_2020-10-13_13-10-10",
            self.example_configs.s3_prefix,
        )

    def test_job_type_default(self):
        """Tests default, valid, and invalid job_type property"""
        self.assertEqual("default", self.example_configs.job_type)

    def test_job_type_validation(self):
        """Test job_type is validated against list context provided."""
        base_configs = self.example_configs.model_dump(
            exclude={"job_type": True, "s3_bucket": True, "s3_prefix": True}
        )
        with validation_context({"job_types": ["default", "ecephys"]}):
            round_trip_model = UploadJobConfigsV2(
                job_type="ecephys",
                **base_configs,
            )
        self.assertEqual("ecephys", round_trip_model.job_type)

    def test_job_type_validation_fail(self):
        """Test job_type is validated against list context provided and
        fails validation."""
        base_configs = self.example_configs.model_dump(
            exclude={"job_type": True, "s3_bucket": True, "s3_prefix": True}
        )
        with self.assertRaises(ValidationError) as err:
            with validation_context({"job_types": ["default", "ecephys"]}):
                UploadJobConfigsV2(
                    job_type="random_string",
                    **base_configs,
                )
        err_msg = json.loads(err.exception.json())[0]["msg"]
        self.assertEqual(
            (
                "Value error, random_string must be one of "
                "['default', 'ecephys']"
            ),
            err_msg,
        )

    def test_project_name_validation(self):
        """Test project_name is validated against list context provided."""
        model = json.loads(self.example_configs.model_dump_json())
        with validation_context(
            {"project_names": ["Behavior Platform", "Other Platform"]}
        ):
            round_trip_model = UploadJobConfigsV2(**model)
        self.assertEqual(
            "behavior_123456_2020-10-13_13-10-10",
            round_trip_model.s3_prefix,
        )

    def test_project_name_validation_fail(self):
        """Test project_name is validated against list context provided and
        fails validation."""
        model = json.loads(self.example_configs.model_dump_json())
        with self.assertRaises(ValidationError) as err:
            with validation_context({"project_names": ["Other Platform"]}):
                UploadJobConfigsV2(**model)
        err_msg = json.loads(err.exception.json())[0]["msg"]
        self.assertEqual(
            (
                "Value error, Behavior Platform must be one of "
                "['Other Platform']"
            ),
            err_msg,
        )

    def test_s3_bucket(self):
        """Test s3_bucket allowed values"""
        default_configs = UploadJobConfigsV2(**self.base_configs)
        base_configs = default_configs.model_dump(
            exclude={"s3_bucket": True, "s3_prefix": True}
        )
        open_configs = UploadJobConfigsV2(s3_bucket="open", **base_configs)
        private_configs = UploadJobConfigsV2(
            s3_bucket="private", **base_configs
        )
        self.assertEqual("default", default_configs.s3_bucket)
        self.assertEqual("open", open_configs.s3_bucket)
        self.assertEqual("private", private_configs.s3_bucket)

    def test_round_trip(self):
        """Tests model can be serialized and de-serialized easily"""
        model_json = self.example_configs.model_dump_json()
        deserialized = UploadJobConfigsV2.model_validate_json(model_json)
        self.assertEqual(self.example_configs, deserialized)

    def test_extra_ignore(self):
        """Tests that extra fields can be passed into model."""
        config = UploadJobConfigsV2(
            **self.base_configs,
            extra_field_1="an extra field",
        )
        config_json = config.model_dump_json()
        self.assertEqual(
            config, UploadJobConfigsV2.model_validate_json(config_json)
        )
        # this also ignores incorrect computed fields
        config_json = config.model_dump(exclude={"s3_prefix": True})
        config = UploadJobConfigsV2(s3_prefix="random_string", **config_json)
        self.assertNotEqual("random_string", config.s3_prefix)
        self.assertEqual(self.example_configs, config)

    def test_check_tasks(self):
        """Tests that tasks can be set correctly"""
        configs = self.base_configs.copy()
        configs["tasks"] = {
            "modality_transformation_settings": {
                "behavior-videos": Task(
                    job_settings={
                        "input_source": "dir/data_set_1",
                        "chunk": "1",
                    }
                ),
                "ecephys": Task(
                    job_settings={
                        "input_source": "dir/data_set_2",
                        "chunk": "1",
                    }
                ),
            },
            "gather_final_metadata": Task(skip_task=True),
            "register_data_asset_to_codeocean": Task(skip_task=True),
        }

        job_configs = UploadJobConfigsV2(**configs)
        self.assertEqual(3, len(job_configs.tasks))
        self.assertEqual(
            2,
            len(job_configs.tasks["modality_transformation_settings"].items()),
        )


class TestSubmitJobRequestV2(unittest.TestCase):
    """Tests SubmitJobRequestV2 class"""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up example configs to be used in tests"""
        example_upload_config = UploadJobConfigsV2(
            job_type="default",
            project_name="Behavior Platform",
            platform=Platform.BEHAVIOR,
            modalities=[
                Modality.BEHAVIOR_VIDEOS,
            ],
            subject_id="123456",
            acq_datetime=datetime(2020, 10, 13, 13, 10, 10),
            tasks={
                "modality_transformation_settings": Task(
                    job_settings={
                        "input_source": "dir/data_set_1",
                    },
                )
            },
        )
        cls.example_upload_config = example_upload_config

    def test_min_length(self):
        """Tests error is raised if no job list is empty"""
        with self.assertRaises(ValidationError) as e:
            SubmitJobRequestV2(upload_jobs=[])
        expected_message = (
            "List should have at least 1 item after validation, not 0"
        )
        actual_message = json.loads(e.exception.json())[0]["msg"]
        self.assertEqual(1, len(json.loads(e.exception.json())))
        self.assertEqual(expected_message, actual_message)

    def test_max_length(self):
        """Tests error is raised if job list is greater than maximum allowed"""
        upload_job = UploadJobConfigsV2(
            **self.example_upload_config.model_dump(round_trip=True)
        )
        with self.assertRaises(ValidationError) as e:
            SubmitJobRequestV2(upload_jobs=[upload_job for _ in range(0, 51)])
        expected_message = (
            "List should have at most 50 items after validation, not 51"
        )
        actual_message = json.loads(e.exception.json())[0]["msg"]
        self.assertEqual(1, len(json.loads(e.exception.json())))
        self.assertEqual(expected_message, actual_message)

    def test_default_settings(self):
        """Tests defaults are set correctly."""
        upload_job = UploadJobConfigsV2(
            **self.example_upload_config.model_dump(round_trip=True)
        )
        job_settings = SubmitJobRequestV2(upload_jobs=[upload_job])
        self.assertIsNone(job_settings.user_email)
        self.assertEqual({"fail"}, job_settings.email_notification_types)
        self.assertEqual("transform_and_upload_v2", job_settings.dag_id)

    def test_non_default_settings(self):
        """Tests user can modify the settings."""
        upload_job_configs = self.example_upload_config.model_dump(
            round_trip=True
        )
        job_settings = SubmitJobRequestV2(
            user_email="abc@acme.com",
            email_notification_types={"begin", "fail"},
            upload_jobs=[UploadJobConfigsV2(**upload_job_configs)],
        )
        self.assertEqual("abc@acme.com", job_settings.user_email)
        self.assertEqual(
            {"begin", "fail"},
            job_settings.email_notification_types,
        )

    def test_email_validation(self):
        """Tests user can not input invalid email address."""
        upload_job_configs = self.example_upload_config.model_dump(
            round_trip=True
        )
        with self.assertRaises(ValidationError) as e:
            SubmitJobRequestV2(
                user_email="some user",
                upload_jobs=[UploadJobConfigsV2(**upload_job_configs)],
            )
        # email_validator changed error message across versions. We can just
        # do a quick check that the error message at least contains this part.
        expected_error_message = "value is not a valid email address: "
        actual_error_message = json.loads(e.exception.json())[0]["msg"]
        # Check only 1 validation error is raised
        self.assertEqual(1, len(json.loads(e.exception.json())))
        self.assertIn(expected_error_message, actual_error_message)

    def test_propagate_email_settings(self):
        """Tests global email settings is propagated to individual jobs."""
        example_job_configs = self.example_upload_config.model_dump(
            exclude={"user_email", "email_notification_types"}, round_trip=True
        )
        new_job = UploadJobConfigsV2(
            user_email="xyz@acme.org",
            email_notification_types=["all"],
            **example_job_configs,
        )
        job_settings = SubmitJobRequestV2(
            user_email="abc@acme.org",
            email_notification_types={"begin", "fail"},
            upload_jobs=[
                new_job,
                UploadJobConfigsV2(**example_job_configs),
            ],
        )
        self.assertEqual(
            "xyz@acme.org", job_settings.upload_jobs[0].user_email
        )
        self.assertEqual(
            "abc@acme.org", job_settings.upload_jobs[1].user_email
        )
        self.assertEqual(
            {"all"}, job_settings.upload_jobs[0].email_notification_types
        )
        self.assertEqual(
            {"begin", "fail"},
            job_settings.upload_jobs[1].email_notification_types,
        )

    def test_check_duplicate_upload_jobs(self):
        """Tests that duplicate upload jobs are not allowed."""
        job_configs = self.example_upload_config.model_dump(
            mode="json", exclude={"subject_id"}
        )
        upload_jobs = [
            UploadJobConfigsV2(**job_configs, subject_id=subject_id)
            for subject_id in ["123456", "123457", "123458", "123456"]
        ]
        with self.assertRaises(ValidationError) as e:
            SubmitJobRequestV2(upload_jobs=upload_jobs)
        errors = json.loads(e.exception.json())
        self.assertEqual(1, len(errors))
        self.assertIn(
            "Duplicate jobs found for behavior_123456_2020-10-13_13-10-10",
            errors[0]["msg"],
        )

    def test_check_duplicate_upload_jobs_same_prefix(self):
        """Tests that upload_jobs with same s3_prefix but different configs
        are allowed"""
        job_configs = self.example_upload_config.model_dump(
            mode="json", exclude={"tasks"}
        )
        upload_jobs = list()
        for i in range(10):
            tasks = {
                "modality_transformation_settings": Task(
                    job_settings={"input_source": "dir/data", "chunk": str(i)}
                )
            }
            upload_jobs.append(UploadJobConfigsV2(**job_configs, tasks=tasks))
        submit_job_request = SubmitJobRequestV2(upload_jobs=upload_jobs)
        self.assertEqual(10, len(submit_job_request.upload_jobs))

    def test_current_jobs_validation(self):
        """Tests job validation against current_jobs provided in context."""
        job_configs = self.example_upload_config.model_dump(
            mode="json", exclude={"subject_id"}
        )
        submitted_job_request = SubmitJobRequestV2(
            upload_jobs=[
                UploadJobConfigsV2(**job_configs, subject_id=subject_id)
                for subject_id in ["123456", "123457", "123458"]
            ]
        )
        current_jobs = [
            j.model_dump(mode="json", exclude_none=True)
            for j in submitted_job_request.upload_jobs
        ]
        new_job = UploadJobConfigsV2(**job_configs, subject_id="123459")
        with validation_context({"current_jobs": current_jobs}):
            submit_job_request = SubmitJobRequestV2(upload_jobs=[new_job])
        self.assertEqual(1, len(submit_job_request.upload_jobs))
        self.assertEqual(
            "behavior_123459_2020-10-13_13-10-10",
            submit_job_request.upload_jobs[0].s3_prefix,
        )

    def test_current_jobs_validation_fail(self):
        """Tests job validation when an upload_job is already running."""
        submitted_job_request = SubmitJobRequestV2(
            upload_jobs=[self.example_upload_config]
        )
        current_jobs_1 = [
            j.model_dump(mode="json", exclude_none=True)
            for j in submitted_job_request.upload_jobs
        ]
        current_jobs_2 = [
            submitted_job_request.model_dump(mode="json", exclude_none=True)
        ]
        for current_jobs in [current_jobs_1, current_jobs_2]:
            with self.assertRaises(ValidationError) as err:
                with validation_context({"current_jobs": current_jobs}):
                    SubmitJobRequestV2(
                        upload_jobs=[self.example_upload_config]
                    )
            err_msg = json.loads(err.exception.json())[0]["msg"]
            self.assertEqual(
                (
                    "Value error, Job is already running/queued for "
                    "behavior_123456_2020-10-13_13-10-10"
                ),
                err_msg,
            )


if __name__ == "__main__":
    unittest.main()



"""Module to handle setting up logger"""

import logging
from typing import Literal, Optional

from logging_loki import LokiHandler
from pydantic import Field
from pydantic_settings import BaseSettings


class LoggingConfigs(BaseSettings):
    """Configs for logger"""

    env_name: Optional[str] = Field(
        default=None, description="Can be used to help tag logging source."
    )
    loki_uri: Optional[str] = Field(
        default=None, description="URI of Loki logging server."
    )
    log_level: Literal[
        "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"
    ] = Field(default="INFO", description="Log level")

    @property
    def app_name(self):
        """Build app name from configs"""
        package_name = __package__
        base_name = package_name.split(".")[0].replace("_", "-")
        app_name = (
            base_name
            if self.env_name is None
            else f"{base_name}-{self.env_name}"
        )
        return app_name

    @property
    def loki_path(self):
        """Full path to log loki messages to"""
        return (
            None
            if self.loki_uri is None
            else f"{self.loki_uri}/loki/api/v1/push"
        )


def get_logger(log_configs: LoggingConfigs) -> logging.Logger:
    """Return a logger that can be used to log messages."""
    level = logging.getLevelName(log_configs.log_level)
    logger = logging.getLogger(__name__)
    logger.setLevel(level)
    if log_configs.loki_uri is not None:
        handler = LokiHandler(
            url=log_configs.loki_path,
            version="1",
            tags={"application": log_configs.app_name},
        )
        logger.addHandler(handler)
    return logger



"""Module to configure and create xlsx job upload template"""

import datetime
from io import BytesIO
from typing import Any, ClassVar, Dict, List

from aind_data_schema_models.modalities import Modality
from aind_data_schema_models.platforms import Platform
from openpyxl import Workbook
from openpyxl.styles import Font
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.datavalidation import DataValidation
from pydantic import BaseModel


class JobUploadTemplate(BaseModel):
    """Class to configure and create xlsx job upload template"""

    FILE_NAME: ClassVar[str] = "job_upload_template.xlsx"
    _NUM_TEMPLATE_ROWS: ClassVar[int] = 20
    _XLSX_DATETIME_FORMAT: ClassVar[str] = "YYYY-MM-DDTHH:mm:ss"
    _HEADERS: ClassVar[List[str]] = [
        "job_type",
        "project_name",
        "platform",
        "acq_datetime",
        "subject_id",
        "metadata_dir",
        "modality0",
        "modality0.input_source",
        "modality1",
        "modality1.input_source",
    ]
    _SAMPLE_JOBS: ClassVar[List[List[Any]]] = [
        [
            "default",
            "Behavior Platform",
            Platform.BEHAVIOR.abbreviation,
            datetime.datetime(2023, 10, 4, 4, 0, 0),
            "123456",
            "/allen/aind/stage/fake/metadata_dir",
            Modality.BEHAVIOR_VIDEOS.abbreviation,
            "/allen/aind/stage/fake/dir",
            Modality.BEHAVIOR.abbreviation,
            "/allen/aind/stage/fake/dir",
        ],
        [
            "default",
            "Ophys Platform - SLAP2",
            Platform.SMARTSPIM.abbreviation,
            datetime.datetime(2023, 3, 4, 16, 30, 0),
            "654321",
            "/allen/aind/stage/fake/Config",
            Modality.SPIM.abbreviation,
            "/allen/aind/stage/fake/dir",
        ],
        [
            "default",
            "Ephys Platform",
            Platform.ECEPHYS.abbreviation,
            datetime.datetime(2023, 1, 30, 19, 1, 0),
            "654321",
            None,
            Modality.ECEPHYS.abbreviation,
            "/allen/aind/stage/fake/dir",
            Modality.BEHAVIOR_VIDEOS.abbreviation,
            "/allen/aind/stage/fake/dir",
        ],
    ]

    @classmethod
    def _get_validators(cls) -> List[Dict[str, Any]]:
        """
        Returns
        -------
        List[Dict[str, Any]]
          A list of validators for fields that require validation.

        """
        return [
            {
                "name": "platform",
                "type": "list",
                "options": list(Platform.abbreviation_map.keys()),
                "column_indexes": [cls._HEADERS.index("platform")],
            },
            {
                "name": "modality",
                "type": "list",
                "options": list(Modality.abbreviation_map.keys()),
                "column_indexes": [
                    cls._HEADERS.index("modality0"),
                    cls._HEADERS.index("modality1"),
                ],
            },
            {
                "name": "datetime",
                "type": "date",
                "column_indexes": [cls._HEADERS.index("acq_datetime")],
            },
        ]

    @classmethod
    def create_excel_sheet_filestream(cls) -> BytesIO:
        """Create job template as xlsx filestream"""
        xl_io = BytesIO()
        workbook = Workbook()
        workbook.iso_dates = True
        worksheet = workbook.active
        worksheet.append(cls._HEADERS)
        for job in cls._SAMPLE_JOBS:
            worksheet.append(job)
        # data validators
        for validator in cls._get_validators():
            dv_type = validator["type"]
            dv_name = validator["name"]
            dv_params = {
                "type": dv_type,
                "promptTitle": dv_name,
                "error": f"Invalid {dv_name}.",
                "allow_blank": True,
                "showErrorMessage": True,
                "showInputMessage": True,
            }
            if dv_type == "list":
                dv_params["formula1"] = f'"{(",").join(validator["options"])}"'
                dv_params["prompt"] = f"Select a {dv_name} from the dropdown"
            elif dv_type == "date":
                dv_params["prompt"] = "Provide a {} using {}".format(
                    dv_name, cls._XLSX_DATETIME_FORMAT
                )
            dv = DataValidation(**dv_params)
            for i in validator["column_indexes"]:
                col = get_column_letter(i + 1)
                col_range = f"{col}2:{col}{cls._NUM_TEMPLATE_ROWS}"
                dv.add(col_range)
                if dv_type != "date":
                    continue
                for (cell,) in worksheet[col_range]:
                    cell.number_format = cls._XLSX_DATETIME_FORMAT
            worksheet.add_data_validation(dv)
        # formatting
        bold = Font(bold=True)
        for cell in worksheet[1]:
            cell.font = bold
            worksheet.column_dimensions[cell.column_letter].auto_size = True
        # save file
        workbook.save(xl_io)
        workbook.close()
        return xl_io



"""Core models for using V2 of aind-data-transfer-service"""

import json
from contextlib import contextmanager
from contextvars import ContextVar
from datetime import datetime
from typing import Any, Dict, List, Literal, Optional, Set, Union

from aind_data_schema_models.data_name_patterns import build_data_name
from aind_data_schema_models.modalities import Modality
from aind_data_schema_models.platforms import Platform
from pydantic import (
    BaseModel,
    ConfigDict,
    EmailStr,
    Field,
    ValidationInfo,
    computed_field,
    field_validator,
    model_validator,
)
from pydantic_settings import BaseSettings

_validation_context: ContextVar[Union[Dict[str, Any], None]] = ContextVar(
    "_validation_context", default=None
)


@contextmanager
def validation_context(context: Union[Dict[str, Any], None]) -> None:
    """
    Following guide in:
    https://docs.pydantic.dev/latest/concepts/validators/#validation-context
    Parameters
    ----------
    context : Union[Dict[str, Any], None]

    Returns
    -------
    None

    """
    token = _validation_context.set(context)
    try:
        yield
    finally:
        _validation_context.reset(token)


class Task(BaseModel):
    """Configuration for a task run during a data transfer upload job."""

    skip_task: bool = Field(
        default=False,
        description=(
            "Skip running this task. If true, only task_id and skip_step are "
            "required."
        ),
        title="Skip Step",
    )
    image: Optional[str] = Field(
        default=None, description="Name of docker image to run", title="Image"
    )
    image_version: Optional[str] = Field(
        default=None,
        description="Version of docker image to run",
        title="Image Version",
    )
    image_resources: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Slurm environment. Must be json serializable.",
        title="Image Resources",
    )
    job_settings: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Settings for the job.",
        title="Job Settings",
    )
    command_script: Optional[str] = Field(
        default=None,
        description=(
            """
            Command script to run. A few strings may be replaced:
            %JOB_SETTINGS: This will be replaced with json.dumps(job_settings)
            %OUTPUT_LOCATION: Output location such as a local directory
            %S3_LOCATION: Location of S3 where to upload data to
            %INPUT_SOURCE: If a job requires a dynamic input source,
             then this may be replaced.
            %IMAGE: The containerized image.
            %IMAGE_VERSION: The image version.
            %ENV_FILE: An environment file location, such as aws configs.
            """
        ),
    )

    @field_validator(
        "image_resources",
        "job_settings",
        mode="after",
    )
    def validate_json_serializable(
        cls, v: Optional[Dict[str, Any]], info: ValidationInfo
    ) -> Optional[Dict[str, Any]]:
        """Validate that fields are json serializable."""
        if v is not None:
            try:
                json.dumps(v)
            except Exception as e:
                raise ValueError(
                    f"{info.field_name} must be json serializable! If "
                    f"converting from a Pydantic model, please use "
                    f'model.model_dump(mode="json"). {e}'
                )
        return v


class UploadJobConfigsV2(BaseSettings):
    """Configuration for a data transfer upload job"""

    # noinspection PyMissingConstructor
    def __init__(self, /, **data: Any) -> None:
        """Add context manager to init for validating fields."""
        self.__pydantic_validator__.validate_python(
            data,
            self_instance=self,
            context=_validation_context.get(),
        )

    model_config = ConfigDict(use_enum_values=True, extra="ignore")

    job_type: str = Field(
        ...,
        description=(
            "Job type for the upload job. Tasks will be run based on the "
            "job_type unless otherwise specified in task_overrides."
        ),
        title="Job Type",
    )

    user_email: Optional[EmailStr] = Field(
        default=None,
        description=(
            "Optional email address to receive job status notifications"
        ),
    )
    email_notification_types: Optional[
        Set[Literal["begin", "end", "fail", "retry", "all"]]
    ] = Field(
        default=None,
        description=(
            "Types of job statuses to receive email notifications about"
        ),
    )
    s3_bucket: Literal["private", "open", "default"] = Field(
        default="default",
        description=(
            "Bucket where data will be uploaded. If not provided, will upload "
            "to default bucket."
        ),
        title="S3 Bucket",
    )

    project_name: str = Field(
        ..., description="Name of project", title="Project Name"
    )
    platform: Platform.ONE_OF = Field(
        ..., description="Platform", title="Platform"
    )
    modalities: List[Modality.ONE_OF] = Field(
        ...,
        description="Data collection modalities",
        title="Modalities",
        min_length=1,
    )
    subject_id: str = Field(..., description="Subject ID", title="Subject ID")
    acq_datetime: datetime = Field(
        ...,
        description="Datetime data was acquired",
        title="Acquisition Datetime",
    )
    tasks: Dict[str, Union[Task, Dict[str, Task]]] = Field(
        ...,
        description=(
            "Dictionary of tasks to run with custom settings. The key must be "
            "the task_id and the value must be the task or list of tasks."
        ),
        title="Tasks",
    )

    @computed_field
    def s3_prefix(self) -> str:
        """Construct s3_prefix from configs."""
        return build_data_name(
            label=f"{self.platform.abbreviation}_{self.subject_id}",
            creation_datetime=self.acq_datetime,
        )

    @field_validator("job_type", "project_name", mode="before")
    def validate_with_context(cls, v: str, info: ValidationInfo) -> str:
        """
        Validate certain fields. If a list of accepted values is provided in a
        context manager, then it will validate against the list. Otherwise, it
        won't raise any validation error.

        Parameters
        ----------
        v : str
          Value input into the field.
        info : ValidationInfo

        Returns
        -------
        str

        """
        valid_list = (info.context or dict()).get(f"{info.field_name}s")
        if valid_list is not None and v not in valid_list:
            raise ValueError(f"{v} must be one of {valid_list}")
        else:
            return v


class SubmitJobRequestV2(BaseSettings):
    """Main request that will be sent to the backend. Bundles jobs into a list
    and allows a user to add an email address to receive notifications."""

    # noinspection PyMissingConstructor
    def __init__(self, /, **data: Any) -> None:
        """Add context manager to init for validating upload_jobs."""
        self.__pydantic_validator__.validate_python(
            data,
            self_instance=self,
            context=_validation_context.get(),
        )

    model_config = ConfigDict(use_enum_values=True, extra="ignore")

    dag_id: Literal["transform_and_upload_v2"] = "transform_and_upload_v2"
    user_email: Optional[EmailStr] = Field(
        default=None,
        description=(
            "Optional email address to receive job status notifications"
        ),
    )
    email_notification_types: Set[
        Literal["begin", "end", "fail", "retry", "all"]
    ] = Field(
        default={"fail"},
        description=(
            "Types of job statuses to receive email notifications about"
        ),
    )
    upload_jobs: List[UploadJobConfigsV2] = Field(
        ...,
        description="List of upload jobs to process. Max of 50 at a time.",
        min_length=1,
        max_length=50,
    )

    @model_validator(mode="after")
    def propagate_email_settings(self):
        """Propagate email settings from global to individual jobs"""
        global_email_user = self.user_email
        global_email_notification_types = self.email_notification_types
        for upload_job in self.upload_jobs:
            if global_email_user is not None and upload_job.user_email is None:
                upload_job.user_email = global_email_user
            if upload_job.email_notification_types is None:
                upload_job.email_notification_types = (
                    global_email_notification_types
                )
        return self

    @model_validator(mode="after")
    def check_duplicate_upload_jobs(self, info: ValidationInfo):
        """Validate that there are no duplicate upload jobs. If a list of
        current jobs is provided in a context manager, jobs are also checked
        against the list."""
        jobs_map = dict()
        # check jobs with the same s3_prefix
        for job in self.upload_jobs:
            prefix = job.s3_prefix
            job_json = json.dumps(
                job.model_dump(mode="json", exclude_none=True), sort_keys=True
            )
            jobs_map.setdefault(prefix, set())
            if job_json in jobs_map[prefix]:
                raise ValueError(f"Duplicate jobs found for {prefix}")
            jobs_map[prefix].add(job_json)
        # check against any jobs in the context
        current_jobs = (info.context or dict()).get("current_jobs", list())
        for job in current_jobs:
            jobs_to_check = job.get("upload_jobs", [job])
            for j in jobs_to_check:
                prefix = j.get("s3_prefix")
                if (
                    prefix is not None
                    and prefix in jobs_map
                    and json.dumps(j, sort_keys=True) in jobs_map[prefix]
                ):
                    raise ValueError(
                        f"Job is already running/queued for {prefix}"
                    )
        return self



"""Module to test hpc client classes"""

import asyncio
import unittest
from unittest.mock import MagicMock, patch

from aind_data_transfer_service.hpc.client import HpcClient, HpcClientConfigs
from aind_data_transfer_service.hpc.models import HpcJobSubmitSettings


class TestHpcClientConfigs(unittest.TestCase):
    """Tests methods in HpcClientConfigs class"""

    def test_hpc_client_configs(self):
        """Tests that attributes are set correctly"""
        hpc_client_configs = HpcClientConfigs(
            hpc_host="hpc_host",
            hpc_port=80,
            hpc_api_endpoint="api/slurm/v0.0.39",
            hpc_username="hpc_username",
            hpc_password="hpc_password",
            hpc_token="hpc_jwt",
        )
        self.assertEqual("hpc_host", hpc_client_configs.hpc_host)
        self.assertEqual(80, hpc_client_configs.hpc_port)
        self.assertEqual(
            "http://hpc_host:80/api/slurm/v0.0.39", hpc_client_configs.hpc_url
        )
        self.assertEqual("hpc_username", hpc_client_configs.hpc_username)
        self.assertEqual(
            "hpc_password", hpc_client_configs.hpc_password.get_secret_value()
        )
        self.assertEqual(
            "hpc_jwt", hpc_client_configs.hpc_token.get_secret_value()
        )


class TestHpcClient(unittest.TestCase):
    """Tests methods in HpcClient class"""

    hpc_client_configs = HpcClientConfigs(
        hpc_host="hpc_host",
        hpc_username="hpc_username",
        hpc_password="hpc_password",
        hpc_token="hpc_jwt",
    )

    def test_properties(self):
        """Tests that the url properties are correct."""
        hpc_client = HpcClient(configs=self.hpc_client_configs)
        self.assertEqual("http://hpc_host/job", hpc_client._job_status_url)
        self.assertEqual("http://hpc_host/nodes", hpc_client._node_status_url)
        self.assertEqual(
            "http://hpc_host/job/submit", hpc_client._job_submit_url
        )

    @patch("httpx.AsyncClient.get")
    def test_node_status_response(self, mock_get: MagicMock):
        """Tests that the node status request is sent correctly."""
        mock_get.return_value = {"message": "A mocked message"}
        hpc_client = HpcClient(configs=self.hpc_client_configs)
        response = asyncio.run(hpc_client.get_node_status())
        self.assertEqual({"message": "A mocked message"}, response)
        mock_get.assert_called_once_with(
            url="http://hpc_host/nodes",
            headers={
                "X-SLURM-USER-NAME": "hpc_username",
                "X-SLURM-USER-PASSWORD": "hpc_password",
                "X-SLURM-USER-TOKEN": "hpc_jwt",
            },
        )

    @patch("httpx.AsyncClient.get")
    def test_job_status_response(self, mock_get: MagicMock):
        """Tests that the job status request is sent correctly"""
        mock_get.return_value = {"message": "A mocked message"}
        hpc_client = HpcClient(configs=self.hpc_client_configs)
        response = asyncio.run(hpc_client.get_job_status(job_id="12345"))
        self.assertEqual({"message": "A mocked message"}, response)
        mock_get.assert_called_once_with(
            url="http://hpc_host/job/12345",
            headers={
                "X-SLURM-USER-NAME": "hpc_username",
                "X-SLURM-USER-PASSWORD": "hpc_password",
                "X-SLURM-USER-TOKEN": "hpc_jwt",
            },
        )

    @patch("httpx.AsyncClient.get")
    def test_get_jobs_response(self, mock_get: MagicMock):
        """Tests that the job status request is sent correctly"""
        mock_get.return_value = {"message": "A mocked message"}
        hpc_client = HpcClient(configs=self.hpc_client_configs)
        response = asyncio.run(hpc_client.get_jobs())
        self.assertEqual({"message": "A mocked message"}, response)
        mock_get.assert_called_once_with(
            url="http://hpc_host/jobs",
            headers={
                "X-SLURM-USER-NAME": "hpc_username",
                "X-SLURM-USER-PASSWORD": "hpc_password",
                "X-SLURM-USER-TOKEN": "hpc_jwt",
            },
        )

    @patch("httpx.AsyncClient.post")
    def test_submit_job_response(self, mock_post: MagicMock):
        """Tests that the job submission request is sent correctly"""
        mock_post.return_value = {"message": "A mocked message"}
        hpc_client = HpcClient(configs=self.hpc_client_configs)
        response = asyncio.run(
            hpc_client.submit_job(job_def={"job": {"some_job"}})
        )
        self.assertEqual({"message": "A mocked message"}, response)
        mock_post.assert_called_once_with(
            url="http://hpc_host/job/submit",
            json={"job": {"some_job"}},
            headers={
                "X-SLURM-USER-NAME": "hpc_username",
                "X-SLURM-USER-PASSWORD": "hpc_password",
                "X-SLURM-USER-TOKEN": "hpc_jwt",
            },
        )

    @patch("httpx.AsyncClient.post")
    def test_submit_hpc_job_response(self, mock_post: MagicMock):
        """Tests that the job submission request is sent correctly"""
        mock_post.return_value = {"message": "A mocked message"}
        hpc_client = HpcClient(configs=self.hpc_client_configs)
        response = asyncio.run(
            hpc_client.submit_hpc_job(
                script="Hello World!",
                job=HpcJobSubmitSettings(name="test_job"),
            )
        )
        self.assertEqual({"message": "A mocked message"}, response)
        mock_post.assert_called_once_with(
            url="http://hpc_host/job/submit",
            json={"script": "Hello World!", "job": {"name": "test_job"}},
            headers={
                "X-SLURM-USER-NAME": "hpc_username",
                "X-SLURM-USER-PASSWORD": "hpc_password",
                "X-SLURM-USER-TOKEN": "hpc_jwt",
            },
        )

    @patch("httpx.AsyncClient.post")
    def test_submit_hpc_jobs_response(self, mock_post: MagicMock):
        """Tests that the jobs submission request is sent correctly"""
        mock_post.return_value = {"message": "A mocked message"}
        hpc_client = HpcClient(configs=self.hpc_client_configs)
        jobs = [
            HpcJobSubmitSettings(name="test_job1"),
            HpcJobSubmitSettings(name="test_job2"),
        ]
        response = asyncio.run(
            hpc_client.submit_hpc_job(script="Hello World!", jobs=jobs)
        )
        self.assertEqual({"message": "A mocked message"}, response)
        mock_post.assert_called_once_with(
            url="http://hpc_host/job/submit",
            json={
                "script": "Hello World!",
                "jobs": [{"name": "test_job1"}, {"name": "test_job2"}],
            },
            headers={
                "X-SLURM-USER-NAME": "hpc_username",
                "X-SLURM-USER-PASSWORD": "hpc_password",
                "X-SLURM-USER-TOKEN": "hpc_jwt",
            },
        )


if __name__ == "__main__":
    unittest.main()
